 language metachop;

#if MOST_RECENT_STL
#include <stack>
#include <vector>
#else
#include <stack.h>
#include <vector.h>
#endif 

static  int GetNumber ( const PTREE &tree )
{
    return APPLY_CLASS(tree, TreeClass, Type());
}

#include "token.h"

// EncodeClass : encoding of a class
// parameters :
//              tree : the tree to be encoded
//              string : recipient string
void PTREE::EncodeClass ( const PTREE &tree, EString &string )
{
    char    *thePt ; // pointer on string
    
    CoarseEncodeInt(GetNumber(tree), string);
    thePt = (char *)(tree.Value());
    CoarseEncodeInt(strlen(thePt), string);
    string += thePt ;
    return ;
}

// CoarseEncode : CoarseEncode a tree
// parameters :
//              tree : the tree to be encoded
//              string : recipient string
void PTREE::CoarseEncode ( PTREE tree, EString &string )
{
    
    int     arity ;  // arity of node
    int     nbNode ; // number of node
    char    *thePt ;    // a pointer
    
start : 
    if ( !tree ) {
        string += (char) - 1 ;
        return ;
    }
    switch ( tree.NumberTree() ) {
        case TERM_TREE : 
            string += '\1';
            CoarseEncodeInt(TERM_TREE, string);
            CoarseEncode(tree [0], string);
            thePt = (char *)(tree.Value());
            CoarseEncodeInt(strlen(thePt), string);
            string += thePt ;
            return ;
        case CLASS_TREE : 
            string += '\1';
            CoarseEncodeInt(CLASS_TREE, string);
            CoarseEncode(tree [0], string);
            EncodeClass(tree, string);
            return ;
        case REF_TREE : 
            string += (char) - 1 ;
            return ;
        default : 
            {
                
                /* a tree node */
                arity = treearity(tree);
                string += (char)arity ;
                CoarseEncodeInt(tree.NumberTree(), string);
                CoarseEncode(tree [0], string);
                
                /* put good language mask */
                int indexSon ; // index of son
                for ( indexSon = 1 ; indexSon < arity ; indexSon++ ) {
                    CoarseEncode(tree [indexSon], string);
                }
                tree = tree [arity];
                goto start ;
            }
    }
}

// DecodeClass : Decode a class 
// parameters :
//              string : string to be decoded
// return : the tree
PTREE PTREE::DecodeClass ( char *&string )
{
    
    PTREE   myTree ;  // return tree
    PTREE   topTree ; // top  tree
    int     nbNode ;  // number of node
    int     length ;  // length of string
    
    nbNode = CoarseDecodeInt(string);
    length = CoarseDecodeInt(string);
    myTree = LMakeString(string, length);
    topTree = PTREE(nbNode, 1);
    topTree.ReplaceTree(1, myTree);
    string += length ;
    return topTree ;
}

// CoarseDecode : CoarseDecode a tree
// parameters :
//              string : input string
// return : the tree
PTREE PTREE::CoarseDecode ( char *&string )
{
    char    arity ;
    PTREE   myTree ;
    int     nbNode ;
    int     length ;
    PTREE   comment ;
    int currPos = -1 ;
#if defined(NO_STACK)
    vector<PTREE> treeStack ;
#elif defined(STL_3)
    stack<PTREE, vector<PTREE> >    treeStack ;
#else 
    stack<vector<PTREE> >   treeStack ;
#endif


start : 
    arity = *string++ ;
    
    // get arity
    if ( arity == (char) - 1 ) 
        myTree = (PTREE)0 ;
    else {
        
        // get node
        nbNode = CoarseDecodeInt(string);
        comment = CoarseDecode(string);
        
        // decode value
        switch ( arity ) {
            case 1 : 
                switch ( nbNode ) {
                    case TERM_TREE : 
                        length = CoarseDecodeInt(string);
                        myTree = LMakeString(string, length);
                        string += length ;
                        myTree.ReplaceTree(0, comment);
                        break ;
                    case CLASS_TREE : 
                        {
                            myTree = DecodeClass(string);
                            myTree.ReplaceTree(0, comment);
                        }
                        break ;
                    default : 
                        myTree = PTREE(nbNode, (int)arity);
                        myTree.ReplaceTree(1, CoarseDecode(string));
                        myTree.ReplaceTree(0, comment);
                        break ;
                }
                break ;
            default : 
                {
                    
                    // construct node
                    myTree = PTREE(nbNode, (int)arity);
                    myTree.ReplaceTree(0, comment);
                    
                    // put sons
                    int indexSon ; // index of son
                    for ( indexSon = 1 ; indexSon < arity ; indexSon++ ) {
                        myTree.ReplaceTree(indexSon, CoarseDecode(string));
                    }
#if defined(NO_STACK)
		    treeStack.reserve(currPos+2);
		    treeStack [++currPos] = myTree ;
#else
                    treeStack.push((const PTREE &) myTree);
#endif
                    goto start ;
                }
        }
    }

    while ( true) {
#if !defined(NO_STACK)
        currPos = treeStack.size();
#endif
        if (!currPos) break;
        PTREE   topNode ;
#if defined(NO_STACK)
	topNode = treeStack[currPos];
        currPos --;
#else
        topNode = treeStack.top();
        treeStack.pop();
#endif
        topNode.ReplaceTree(topNode.TreeArity(), myTree);
        myTree = topNode ;
    }

    return myTree ;
}
